#include <stdio.h>
#include <pcap/pcap.h>

#include "jni.h"
#include "pcksniff_Sniffer.h"

#define ERR_BUFF_SIZE 512


jobject  jo_consumer ;

JNIEnv  *ev ;

void dispatcher(u_char *c, struct pcap_pkthdr  *header, u_char *pkt_data)
{
  jlong	sec	= header->ts.tv_sec ;
  jlong millis  = header->ts.tv_usec ;

  int pckSize = header->len ;

  jclass    cls = (*ev)->GetObjectClass(ev, jo_consumer) ;
  jmethodID mid = (*ev)->GetMethodID(ev, cls, "consume", "(JJ[B)V");

  jbyteArray packet = (*ev)->NewByteArray(ev,pckSize) ;

  (*ev)->SetByteArrayRegion(ev, packet, 0 , pckSize, pkt_data  );
  (*ev)->CallVoidMethod(ev, jo_consumer, mid, sec, millis, packet);

}

JNIEXPORT jbyteArray JNICALL Java_pcksniff_Sniffer_pcapOpenLive(JNIEnv *env, jobject jobj, jstring js_device, jint snaplen, jint promisc, jint to_ms, jobject jo_ebuf)
{ //
  jboolean        str_copy ;
  char 	         *device = (*env)->GetStringUTFChars(env, js_device, &str_copy) ;
  char		 *ebuf ;
  jbyteArray      ret  ;
  pcap_t	 *pcap ;
  jclass          cls  ;
  jmethodID       mid  ;

  ebuf = (char*)malloc(ERR_BUFF_SIZE) ;

  pcap = pcap_open_live(device,snaplen,promisc,to_ms,ebuf) ;

  if (!pcap)
  {
    ret  = (*env)->NewByteArray(env,0) ;
    cls  = (*env)->GetObjectClass(env,jo_ebuf) ;

    mid = (*env)->GetMethodID(env, cls, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");

    (*env)->CallObjectMethod(env, jo_ebuf, mid, (*env)->NewStringUTF(env,ebuf));

  }
  else
  {
    ret = (*env)->NewByteArray(env,sizeof(void*)) ;
    (*env)->SetByteArrayRegion(env, ret, 0 , sizeof(void*),(char*) &pcap  );
  }

  free(ebuf)   ;
  return ret   ;
}

JNIEXPORT jbyteArray JNICALL Java_pcksniff_Sniffer_pcapOpenOffline(JNIEnv *env, jobject jobj, jstring js_fname, jobject jo_ebuf)
{ //
  char 		 *fname = (*env)->GetStringUTFChars(env, js_fname, 0) ;
  char		 *ebuf ;
  jbyteArray      ret  ;
  pcap_t	 *pcap ;
  jclass          cls  ;
  jmethodID       mid  ;


  ebuf = (char*)malloc(ERR_BUFF_SIZE) ;

  pcap = pcap_open_offline(fname,ebuf) ;

  if (!pcap)
  {
    ret = (*env)->NewByteArray(env,0) ;
    cls = (*env)->GetObjectClass(env, jo_ebuf);
    mid = (*env)->GetMethodID(env, cls, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
    (*env)->CallObjectMethod(env, jo_ebuf, mid, (*env)->NewStringUTF(env,ebuf));
  }
  else
  {
    ret = (*env)->NewByteArray(env,sizeof(void*)) ;
    (*env)->SetByteArrayRegion(env, ret, 0 , sizeof(void*), &pcap  );
  }

  free(ebuf)   ;
  return ret   ;
}

JNIEXPORT jbyteArray JNICALL Java_pcksniff_Sniffer_pcapDumpOpen(JNIEnv *env, jobject jobj, jbyteArray jb_pcap, jstring js_fname)
{ //
  char 		    *fname = (*env)->GetStringUTFChars(env, js_fname, 0) ;
  jbyteArray     ret  ;
  pcap_t	    *pcap ;
  pcap_dumper_t *pcapdumper ;

  (*env)->GetByteArrayRegion(env, jb_pcap , 0 , sizeof(void*), &pcap   );

  pcapdumper = pcap_dump_open(pcap,fname) ;

  if (!pcapdumper)
  {
	ret   = (*env)->NewByteArray(env,0) ;
  }
  else
  {
    ret = (*env)->NewByteArray(env,sizeof(void*)) ;
    (*env)->SetByteArrayRegion(env, ret, 0 , sizeof(void*), &pcapdumper  );
  }

  free(fname) ;

  return ret   ;

}

JNIEXPORT jstring JNICALL Java_pcksniff_Sniffer_pcapLookupDev(JNIEnv *env, jobject jobj, jobject jo_ebuf)
{
  //
  jstring     ret  ;
  char	      *dev ;
  char        *ebuf ;

  ebuf = malloc(ERR_BUFF_SIZE) ;
  dev = pcap_lookupdev(ebuf) ;

  if (!dev)
  { jclass    cls = (*env)->GetObjectClass(env, jo_ebuf);
    jmethodID mid = (*env)->GetMethodID(env, cls, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
    (*env)->CallObjectMethod(env, jo_ebuf, mid, (*env)->NewStringUTF(env,ebuf));
	//
	free(ebuf)   ;
    return (*env)->NewStringUTF(env,"");
  }

  if ( *(dev+1)==0 )
    { char *c = dev ;
      char *d = dev ;
	  while(*d!=0)
      { *c = *d ;
        c++ ;
        d+=2 ;
      }
    }

  free(ebuf)   ;
  return (*env)->NewStringUTF(env, dev);
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapLookupNetNumber(JNIEnv *env, jobject jobj, jstring js_device,jobject jo_ebuf)
{
  char	 	*device = (*env)->GetStringUTFChars(env, js_device, 0) ;
  bpf_u_int32 	netp   ;
  bpf_u_int32   maskp  ;
  char 		*ebuf ;
  int 		result ;

  ebuf = malloc(ERR_BUFF_SIZE) ;

  result = pcap_lookupnet(device,&netp,&maskp,ebuf) ;

  if (result==-1)
  {
     jclass    cls = (*env)->GetObjectClass(env, jo_ebuf);
     jmethodID mid = (*env)->GetMethodID(env, cls, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
     (*env)->CallObjectMethod(env, jo_ebuf, mid, (*env)->NewStringUTF(env,ebuf));
     return 0 ;
  }
  else
  {  return netp ;
  }
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapLookupNetMask(JNIEnv *env, jobject jobj, jstring js_device,jobject jo_ebuf)
{
  char	 		*device = (*env)->GetStringUTFChars(env, js_device, 0) ;
  bpf_u_int32 	netp   ;
  bpf_u_int32   maskp  ;
  char 			*ebuf ;
  int 			 result ;

  ebuf = malloc(ERR_BUFF_SIZE) ;

  result = pcap_lookupnet(device,&netp,&maskp,ebuf) ;


  if (result==-1)
  {
     jclass    cls = (*env)->GetObjectClass(env, jo_ebuf);
     jmethodID mid = (*env)->GetMethodID(env, cls, "append", "(Ljava/lang/String;)Ljava/lang/StringBuffer;");
     (*env)->CallObjectMethod(env, jo_ebuf, mid, (*env)->NewStringUTF(env,ebuf));
     return 0 ;
  }
  else
  {  return maskp ;
  }
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapDispatch(JNIEnv *env, jobject jobj, jbyteArray jb_pcap, jobject jo_cons, jint ji_cnt)
{
  pcap_t *pcap ;

  jo_consumer = jo_cons ;
  ev = env ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );

  return pcap_dispatch(pcap,ji_cnt,dispatcher,NULL) ;
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapLoop(JNIEnv *env, jobject jobj, jbyteArray jb_pcap, jobject jo_cons, jint ji_cnt)
{
  pcap_t *pcap ;
  jo_consumer = jo_cons ;
  ev = env ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );

  return pcap_loop(pcap,ji_cnt,dispatcher,NULL) ;
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapDispatchDump(JNIEnv *env, jobject jobj, jbyteArray jb_pcap,jbyteArray jb_dumper, jint ji_cnt)
{
  pcap_t	    *pcap ;
  pcap_dumper_t *dumper ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );
  (*env)->GetByteArrayRegion(env,jb_dumper, 0 , sizeof(void*), &dumper  );

  return pcap_dispatch(pcap,ji_cnt,pcap_dump,dumper) ;

}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapLoopDump(JNIEnv *env, jobject jobj, jbyteArray jb_pcap,jbyteArray jb_dumper, jint ji_cnt)
{
  pcap_t        *pcap ;
  pcap_dumper_t *dumper ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );
  (*env)->GetByteArrayRegion(env,jb_dumper, 0 , sizeof(void*), &dumper  );

  return pcap_loop(pcap,ji_cnt,pcap_dump,dumper) ;
}

JNIEXPORT jbyteArray JNICALL Java_pcksniff_Sniffer_pcapAllocateBpfProgram(JNIEnv *env, jobject jobj)
{
  struct bpf_program *fp ;

  jbyteArray ret ;

  fp = malloc(sizeof(struct bpf_program)) ;

  ret = (*env)->NewByteArray(env,sizeof(void*)) ;

  (*env)->SetByteArrayRegion(env, ret, 0 , sizeof(void*), &fp  );

  return ret ;
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapCompile(JNIEnv *env, jobject jobj, jbyteArray jb_pcap, jbyteArray jb_bpf_prog, jstring js_query, jint ji_optimize, jint ji_netmask)
{
  pcap_t   				*pcap ;
  struct bpf_program 	*fp ;
  char	   				*query  ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );
  (*env)->GetByteArrayRegion(env,jb_bpf_prog, 0 , sizeof(void*), &fp  );
  query = (*env)->GetStringUTFChars(env, js_query, 0) ;

  return pcap_compile(pcap,fp,query,ji_optimize,ji_netmask) ;
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapSetFilter(JNIEnv *env, jobject jobj, jbyteArray jb_pcap, jbyteArray jb_bpf_prog)
{
  pcap_t   				*pcap ;
  struct bpf_program 	*fp ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );
  (*env)->GetByteArrayRegion(env,jb_bpf_prog, 0 , sizeof(void*), &fp  );

  return pcap_setfilter(pcap,fp) ;
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapStatsPacketsDropped(JNIEnv *env, jobject jobj, jbyteArray jb_pcap)
{
  pcap_t	    *pcap ;
  struct pcap_stat   ps ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );

  if (pcap_stats(pcap,&ps)==-1)
  { return -1 ;
  }
  else
  { return (ps.ps_drop) ;
  }
}

JNIEXPORT jint JNICALL Java_pcksniff_Sniffer_pcapStatsPacketsReceived(JNIEnv *env, jobject jobj, jbyteArray jb_pcap)
{
  pcap_t	    *pcap ;
  struct pcap_stat   ps ;

  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );

  if (pcap_stats(pcap,&ps)==-1)
  { return -1 ;
  }
  else
  { return (ps.ps_recv) ;
  }
}

JNIEXPORT void JNICALL Java_pcksniff_Sniffer_pcapClose(JNIEnv *env, jobject jobj, jbyteArray jb_pcap)
{
  pcap_t			*pcap ;
  (*env)->GetByteArrayRegion(env,jb_pcap, 0 , sizeof(void*), &pcap  );
  pcap_close(pcap) ;
}

JNIEXPORT void JNICALL Java_pcksniff_Sniffer_pcapDumpClose(JNIEnv *env, jobject jobj, jbyteArray jb_dmp)
{
  pcap_dumper_t	*dmp ;
  (*env)->GetByteArrayRegion(env,jb_dmp, 0 , sizeof(void*), &dmp  );
  pcap_dump_close(dmp) ;
}
