package pcksniff;

/**
* The PacketConsumer interface for elaborating the packets received. The class that is interested 
* in processing a packet implements this interface. The object created with that class is passed as parameter 
* to the Sniffer.dispatch() or Sniffer.loop() method. <BR> When a packet is received, 
* the consume method is invoked. 
* 
* @see Sniffer
*/
public interface PacketConsumer
{ 
  /**
  * Invoked when a packet is received.
  * 
  * @param sec  The received time of the packet. 
  * @param millis  The millisecond of received time of the packet.
  * @param packetData The raw data of the packet received. 
  * <BR>
  * <BR>
  * The time is expressed in seconds since January 1, 1970, 00:00:00 GMT.
  */
  public abstract void consume(long sec,long millis, byte[] packetData);
} 